"""
    Integrate module
    ================

    Module with the function to integrate the ordinary differential equations

    .. math:: \\dot{\\boldsymbol{x}} = \\boldsymbol{f}(t, \\boldsymbol{x})

    of the model and its linearized version.

    Description of the module functions
    -----------------------------------

    Two main functions:

    * :obj:`integrate_runge_kutta`
    * :obj:`integrate_runge_kutta_tgls`

"""


# TODO : - test the provided functions with try before proceeding

from numba import njit
import numpy as np
from qgs.functions.util import reverse


def integrate_runge_kutta(f, t0, t, dt, ic=None, forward=True, write_steps=1, b=None, c=None, a=None):
    """
    Integrate the ordinary differential equations (ODEs)

    .. math:: \\dot{\\boldsymbol{x}} = \\boldsymbol{f}(t, \\boldsymbol{x})

    with a specified `Runge-Kutta method`_. The function :math:`\\boldsymbol{f}` should
    be a `Numba`_ jitted function. This function must have a signature ``f(t, x)`` where ``x`` is
    the state value and ``t`` is the time.

    .. _Runge-Kutta method: https://en.wikipedia.org/wiki/Runge%E2%80%93Kutta_methods
    .. _Numba: https://numba.pydata.org/

    Parameters
    ----------
    f: callable
        The `Numba`_-jitted function :math:`\\boldsymbol{f}`.
        Should have the signature``f(t, x)`` where ``x`` is the state value and ``t`` is the time.
    t0: float
        Initial time of the time integration. Corresponds to the initial condition.
        Important if the ODEs are non-autonomous.
    t: float
        Final time of the time integration. Corresponds to the final condition.
        Important if the ODEs are non-autonomous.
    dt: float
        Timestep of the integration.
    ic: None or ~numpy.ndarray(float), optional
        Initial (or final) conditions of the system. Can be a 1D or a 2D array:

        * 1D: Provide a single initial condition.
          Should be of shape (`n_dim`,) where `n_dim` = :math:`\\mathrm{dim}(\\boldsymbol{x})`.
        * 2D: Provide an ensemble of initial condition.
          Should be of shape (`n_traj`, `n_dim`) where `n_dim` = :math:`\\mathrm{dim}(\\boldsymbol{x})`,
          and where `n_traj` is the number of initial conditions.

        If `None`, use a zero initial condition. Default to `None`.
        If the `forward` argument is `False`, it specifies final conditions.
    forward: bool, optional
        Whether to integrate the ODEs forward or backward in time. In case of backward integration, the
        initial condition `ic` becomes a final condition. Default to forward integration.
    write_steps: int, optional
        Save the state of the integration in memory every `write_steps` steps. The other intermediary
        steps are lost. It determines the size of the returned objects. Default is 1.
        Set to 0 to return only the final state.
    b: None or ~numpy.ndarray, optional
        Vector of coefficients :math:`b_i` of the `Runge-Kutta method`_ .
        If `None`, use the classic RK4 method coefficients. Default to `None`.
    c: None or ~numpy.ndarray, optional
        Matrix of coefficients :math:`c_{i,j}` of the `Runge-Kutta method`_ .
        If `None`, use the classic RK4 method coefficients. Default to `None`.
    a: None or ~numpy.ndarray, optional
        Vector of coefficients :math:`a_i` of the `Runge-Kutta method`_ .
        If `None`, use the classic RK4 method coefficients. Default to `None`.

    Returns
    -------
    time, traj: ~numpy.ndarray
        The result of the integration:

        * **time:** Time at which the state of the system was saved. Array of shape (`n_step`,) where
          `n_step` is the number of saved states of the integration.
        * **traj:** Saved dynamical system states. 3D array of shape (`n_traj`, `n_dim`, `n_steps`). If `n_traj` = 1,
          a 2D array of shape (`n_dim`, `n_steps`) is returned instead.

    Examples
    --------

    >>> from numba import njit
    >>> import numpy as np
    >>> from qgs.integrators.integrate import integrate_runge_kutta
    >>> a = 0.25
    >>> F = 16.
    >>> G = 3.
    >>> b = 6.
    >>> # Lorenz 84 example
    >>> @njit
    ... def fL84(t, x):
    ...     xx = -x[1] ** 2 - x[2] ** 2 - a * x[0] + a * F
    ...     yy = x[0] * x[1] - b * x[0] * x[2] - x[1] + G
    ...     zz = b * x[0] * x[1] + x[0] * x[2] - x[2]
    ...     return np.array([xx, yy, zz])
    >>> # no ic
    >>> # write_steps is 1 by default
    >>> tt, traj = integrate_runge_kutta(fL84, t0=0., t=10., dt=0.1)  # 101 steps
    >>> print(traj.shape)
    (3, 101)
    >>> # 1 ic
    >>> ic = 0.1 * np.random.randn(3)
    >>> tt, traj = integrate_runge_kutta(fL84, t0=0., t=10., dt=0.1, ic=ic)  # 101 steps
    >>> print(ic.shape)
    (3,)
    >>> print(traj.shape)
    (3, 101)
    >>> # 4 ic
    >>> ic = 0.1 * np.random.randn(4, 3)
    >>> tt, traj = integrate_runge_kutta(fL84, t0=0., t=10., dt=0.1, ic=ic)  # 101 steps
    >>> print(ic.shape)
    (4, 3)
    >>> print(traj.shape)
    (4, 3, 101)
    """

    if ic is None:
        i = 1
        while True:
            ic = np.zeros(i)
            try:
                x = f(0., ic)
            except:
                i += 1
            else:
                break

        i = len(f(0., ic))
        ic = np.zeros(i)

    if len(ic.shape) == 1:
        ic = ic.reshape((1, -1))

    # Default is RK4
    if a is None and b is None and c is None:
        c = np.array([0., 0.5, 0.5, 1.])
        b = np.array([1./6, 1./3, 1./3, 1./6])
        a = np.zeros((len(c), len(b)))
        a[1, 0] = 0.5
        a[2, 1] = 0.5
        a[3, 2] = 1.

    if forward:
        time_direction = 1
    else:
        time_direction = -1

    time = np.concatenate((np.arange(t0, t, dt), np.full((1,), t)))

    recorded_traj = _integrate_runge_kutta_jit(f, time, ic, time_direction, write_steps, b, c, a)

    if write_steps > 0:
        if forward:
            if time[::write_steps][-1] == time[-1]:
                return time[::write_steps], np.squeeze(recorded_traj)
            else:
                return np.concatenate((time[::write_steps], np.full((1,), t))), np.squeeze(recorded_traj)
        else:
            rtime = reverse(time[::-write_steps])
            if rtime[0] == time[0]:
                return rtime, np.squeeze(recorded_traj)
            else:
                return np.concatenate((np.full((1,), t0), rtime)), np.squeeze(recorded_traj)
    else:
        return time[-1], np.squeeze(recorded_traj)


@njit
def _integrate_runge_kutta_jit(f, time, ic, time_direction, write_steps, b, c, a):

    n_traj = ic.shape[0]
    n_dim = ic.shape[1]

    s = len(b)

    if write_steps == 0:
        n_records = 1
    else:
        tot = time[::write_steps]
        n_records = len(tot)
        if tot[-1] != time[-1]:
            n_records += 1

    recorded_traj = np.zeros((n_traj, n_dim, n_records))
    if time_direction == -1:
        directed_time = reverse(time)
    else:
        directed_time = time

    for i_traj in range(n_traj):
        y = ic[i_traj].copy()
        k = np.zeros((s, n_dim))
        iw = 0
        for ti, (tt, dt) in enumerate(zip(directed_time[:-1], np.diff(directed_time))):

            if write_steps > 0 and np.mod(ti, write_steps) == 0:
                recorded_traj[i_traj, :, iw] = y
                iw += 1

            k.fill(0.)
            for i in range(s):
                y_s = y + dt * a[i] @ k
                k[i] = f(tt + c[i] * dt, y_s)
            y_new = y + dt * b @ k
            y = y_new

        recorded_traj[i_traj, :, -1] = y

    return recorded_traj[:, :, ::time_direction]


@njit
def _tangent_linear_system(fjac, t, xs, x, adjoint):
    if adjoint:
        return fjac(t, xs).transpose() @ x
    else:
        return fjac(t, xs) @ x


# a function that return always zero
@njit
def _zeros_func(t, x):
    return np.zeros_like(x)


def integrate_runge_kutta_tgls(f, fjac, t0, t, dt, ic=None, tg_ic=None,
                               forward=True, adjoint=False, inverse=False, boundary=None,
                               write_steps=1, b=None, c=None, a=None):
    """Integrate simultaneously the ordinary differential equations (ODEs)

    .. math:: \dot{\\boldsymbol{x}} = \\boldsymbol{f}(t, \\boldsymbol{x})

    and its tangent linear model, i.e. the linearized ODEs

    .. math :: \dot{\\boldsymbol{\delta x}} = \\boldsymbol{\mathrm{J}}(t, \\boldsymbol{x}) \cdot \\boldsymbol{\delta x}

    where :math:`\\boldsymbol{\mathrm{J}} = \\frac{\partial \\boldsymbol{f}}{\partial \\boldsymbol{x}}` is the
    Jacobian matrix of :math:`\\boldsymbol{f}`, with a specified `Runge-Kutta method`_.
    To solve this equation, one has to integrate simultaneously both ODEs.

    The function :math:`\\boldsymbol{f}` and :math:`\\boldsymbol{J}` should
    be `Numba`_ jitted functions. These functions must have a signature ``f(t, x)`` and ``J(t, x)`` where ``x`` is
    the state value and ``t`` is the time.

    .. _Runge-Kutta method: https://en.wikipedia.org/wiki/Runge%E2%80%93Kutta_methods
    .. _Numba: https://numba.pydata.org/
    .. _fundamental matrix of solutions: https://en.wikipedia.org/wiki/Fundamental_matrix_(linear_differential_equation)

    Parameters
    ----------
    f: callable
        The `Numba`_-jitted function :math:`\\boldsymbol{f}`.
        Should have the signature``f(t, x)`` where ``x`` is the state value and ``t`` is the time.
    fjac: callable
        The `Numba`_-jitted Jacobian :math:`\\boldsymbol{J}`.
        Should have the signature``J(t, x)`` where ``x`` is the state value and ``t`` is the time.
    t0: float
        Initial time of the time integration. Corresponds to the initial conditions.
    t: float
        Final time of the time integration. Corresponds to the final conditions.
    dt: float
        Timestep of the integration.
    ic: None or ~numpy.ndarray(float), optional
        Initial (or final) conditions of the ODEs :math:`\dot{\\boldsymbol{x}} = \\boldsymbol{f}(t, \\boldsymbol{x})`.
        Can be a 1D or a 2D array:

        * 1D: Provide a single initial condition.
          Should be of shape (`n_dim`,) where `n_dim` = :math:`\mathrm{dim}(\\boldsymbol{x})`.
        * 2D: Provide an ensemble of initial condition.
          Should be of shape (`n_traj`, `n_dim`) where `n_dim` = :math:`\mathrm{dim}(\\boldsymbol{x})`,
          and where `n_traj` is the number of initial conditions.

        If `None`, use a zero initial condition. Default to `None`.
        If the `forward` argument is `False`, it specifies final conditions.
    tg_ic: None or ~numpy.ndarray(float), optional
        Initial (or final) conditions of the linear ODEs
        :math:`\dot{\\boldsymbol{\delta x}} = \\boldsymbol{\mathrm{J}}(t, \\boldsymbol{x}) \cdot \\boldsymbol{\delta x}`. \n
        Can be a 1D, a 2D or a 3D array:

        * 1D: Provide a single initial condition. This initial condition of the linear ODEs will be the same used for each
          initial condition `ic` of the ODEs :math:`\dot{\\boldsymbol{x}} = \\boldsymbol{f}(t, \\boldsymbol{x})`
          Should be of shape (`n_dim`,) where `n_dim` = :math:`\mathrm{dim}(\\boldsymbol{x})`.
        * 2D: Two sub-cases:

            + If `tg_ic.shape[0]`=`ic.shape[0]`, assumes that each initial condition `ic[i]` of :math:`\dot{\\boldsymbol{x}} = \\boldsymbol{f}(t, \\boldsymbol{x})`,
              correspond to a different initial condition `tg_ic[i]`.
            + Else, assumes and integrate an ensemble of `n_tg_traj` initial condition of the linear ODEs for each
              initial condition of :math:`\dot{\\boldsymbol{x}} = \\boldsymbol{f}(t, \\boldsymbol{x})`.
        * 3D: An array of shape (`n_traj`, `n_tg_traj`, `n_dim`) which provide an ensemble of `n_tg_ic` initial conditions
          specific to each of the `n_traj` initial conditions of :math:`\dot{\\boldsymbol{x}} = \\boldsymbol{f}(t, \\boldsymbol{x})`.

        If `None`, use the identity matrix as initial condition, returning the `fundamental matrix of solutions`_ of the
        linear ODEs.
        Default to `None`.
        If the `forward` argument is `False`, it specifies final conditions.
    forward: bool, optional
        Whether to integrate the ODEs forward or backward in time. In case of backward integration, the
        initial condition `ic` becomes a final condition. Default to forward integration.
    adjoint: bool, optional
        Wheter to integrate the tangent :math:`\dot{\\boldsymbol{\delta x}} = \\boldsymbol{\mathrm{J}}(t, \\boldsymbol{x}) \cdot \\boldsymbol{\delta x}`
        or the adjoint linear model :math:`\dot{\\boldsymbol{\delta x}} = \\boldsymbol{\mathrm{J}}^T(t, \\boldsymbol{x}) \cdot \\boldsymbol{\delta x}`.
        Integrate the tangent model by default.
    inverse: bool, optional
        Wheter or not to invert the Jacobian matrix
        :math:`\\boldsymbol{\mathrm{J}}(t, \\boldsymbol{x}) \\rightarrow \\boldsymbol{\mathrm{J}}^{-1}(t, \\boldsymbol{x})`.
        `False` by default.
    boundary: None or callable, optional
        Allow to add a inhomogeneous term to linear ODEs:
        :math:`\dot{\\boldsymbol{\delta x}} = \\boldsymbol{\mathrm{J}}(t, \\boldsymbol{x}) \cdot \\boldsymbol{\delta x} + \Psi(t, \\boldsymbol{x})`.
        The boundary :math:`\Psi` should have the same signature as :math:`\\boldsymbol{\mathrm{J}}`, i.e.  ``func(t, x)``.
        If `None`, don't add anything (homogeneous case). `None` by default.
    write_steps: int, optional
        Save the state of the integration in memory every `write_steps` steps. The other intermediary
        steps are lost. It determines the size of the returned objects. Default is 1.
        Set to 0 to return only the final state.
    b: None or ~numpy.ndarray, optional
        Vector of coefficients :math:`b_i` of the `Runge-Kutta method`_ .
        If `None`, use the classic RK4 method coefficients. Default to `None`.
    c: None or ~numpy.ndarray, optional
        Matrix of coefficients :math:`c_{i,j}` of the `Runge-Kutta method`_ .
        If `None`, use the classic RK4 method coefficients. Default to `None`.
    a: None or ~numpy.ndarray, optional
        Vector of coefficients :math:`a_i` of the `Runge-Kutta method`_ .
        If `None`, use the classic RK4 method coefficients. Default to `None`.

    Returns
    -------
    time, traj, tg_traj: ~numpy.ndarray
        The result of the integration:

        * **time:** Time at which the state of the system was saved. Array of shape (`n_step`,) where
          `n_step` is the number of saved states of the integration.
        * **traj:** Saved states of the ODEs. 3D array of shape (`n_traj`, `n_dim`, `n_steps`). If `n_traj` = 1,
          a 2D array of shape (`n_dim`, `n_steps`) is returned instead.
        * **tg_traj:** Saved states of the linear ODEs.
          Depending on the input initial conditions of both ODEs,
          it is at maximum a 4D array of shape (`n_traj`, `n_tg_traj `n_dim`, `n_steps`).
          If one of the dimension is 1, it is squeezed.


    Examples
    --------

    >>> from numba import njit
    >>> import numpy as np
    >>> from qgs.integrators.integrate import integrate_runge_kutta_tgls
    >>> a = 0.25
    >>> F = 16.
    >>> G = 3.
    >>> b = 6.
    >>> # Lorenz 84 example
    >>> @njit
    ... def fL84(t, x):
    ...     xx = -x[1] ** 2 - x[2] ** 2 - a * x[0] + a * F
    ...     yy = x[0] * x[1] - b * x[0] * x[2] - x[1] + G
    ...     zz = b * x[0] * x[1] + x[0] * x[2] - x[2]
    ...     return np.array([xx, yy, zz])
    >>> @njit
    ... def DfL84(t, x):
    ...    return np.array([[     -a        , -2. * x[1], -2. * x[2]],
    ...                      [x[1] - b * x[2], -1. + x[0], -b * x[0]],
    ...                      [b * x[1] + x[2],  b * x[0], -1. + x[0]]])
    >>> # 4 ic, no tg_ic (fundamental matrix computation of an ensemble of ic)
    >>> ic = 0.1 * np.random.randn(4, 3)
    >>> tt, traj, dtraj = integrate_runge_kutta_tgls(fL84, DfL84, t0=0., t=10., dt=0.1,
    ...                                              ic=ic, write_steps=1) # 101 steps
    >>> print(ic.shape)
    (4, 3)
    >>> print(traj.shape)
    (4, 3, 101)
    >>> print(dtraj.shape)
    (4, 3, 3, 101)
    >>> # 1 ic, 1 tg_ic (one ic and its tg_ic)
    >>> ic = 0.1 * np.random.randn(3)
    >>> tg_ic = 0.001 * np.random.randn(3)
    >>> tt, traj, dtraj = integrate_runge_kutta_tgls(fL84, DfL84, t0=0., t=10., dt=0.1,
    ...                                              ic=ic, tg_ic=tg_ic) # 101 steps
    >>> print(ic.shape)
    (3,)
    >>> print(tg_ic.shape)
    (3,)
    >>> print(traj.shape)
    (3, 101)
    >>> print(dtraj.shape)
    (3, 101)
    >>> # 4 ic, 1 same tg_ic (an ensemble of ic with the same tg_ic)
    >>> ic = 0.1 * np.random.randn(4, 3)
    >>> tt, traj, dtraj = integrate_runge_kutta_tgls(fL84, DfL84, t0=0., t=10., dt=0.1,
    ...                                              ic=ic, tg_ic=tg_ic) # 101 steps
    >>> print(ic.shape)
    (4, 3)
    >>> print(tg_ic.shape)
    (3,)
    >>> print(traj.shape)
    (4, 3, 101)
    >>> print(dtraj.shape)
    (4, 3, 101)
    >>> # 1 ic, 4 tg_ic (an ic with an ensemble of tg_ic in its tangent space)
    >>> ic = 0.1 * np.random.randn(3)
    >>> tg_ic = 0.001 * np.random.randn(4, 3)
    >>> tt, traj, dtraj = integrate_runge_kutta_tgls(fL84, DfL84, t0=0., t=10., dt=0.1,
    ...                                              ic=ic, tg_ic=tg_ic) # 101 steps
    >>> print(ic.shape)
    (3,)
    >>> print(tg_ic.shape)
    (4, 3)
    >>> print(traj.shape)
    (3, 101)
    >>> print(dtraj.shape)
    (4, 3, 101)
    >>> # 2 ic, same 4 tg_ic (an ensemble of 2 ic, both with the same ensemble
    >>> # of tg_ic in their tangent space)
    >>> ic = 0.1 * np.random.randn(2, 3)
    >>> tg_ic = 0.001 * np.random.randn(4, 3)
    >>> tt, traj, dtraj = integrate_runge_kutta_tgls(fL84, DfL84, t0=0., t=10., dt=0.1,
    ...                                              ic=ic, tg_ic=tg_ic) # 101 steps
    >>> print(ic.shape)
    (2, 3)
    >>> print(tg_ic.shape)
    (4, 3)
    >>> print(traj.shape)
    (2, 3, 101)
    >>> print(dtraj.shape)
    (2, 4, 3, 101)
    >>> # 2 ic, 4 different tg_ic (an ensemble of 2 ic, with different ensemble
    >>> # of tg_ic in their tangent space)
    >>> ic = 0.1 * np.random.randn(2, 3)
    >>> tg_ic = 0.001 * np.random.randn(2, 4, 3)
    >>> tt, traj, dtraj = integrate_runge_kutta_tgls(fL84, DfL84, t0=0., t=10., dt=0.1,
    ...                                              ic=ic, tg_ic=tg_ic) # 101 steps
    >>> print(ic.shape)
    (2, 3)
    >>> print(tg_ic.shape)
    (2, 4, 3)
    >>> print(traj.shape)
    (2, 3, 101)
    >>> print(dtraj.shape)
    (2, 4, 3, 101)

    """

    if ic is None:
        i = 1
        while True:
            ic = np.zeros(i)
            try:
                x = f(0., ic)
            except:
                i += 1
            else:
                break

        i = len(f(0., ic))
        ic = np.zeros(i)

    if len(ic.shape) == 1:
        ic = ic.reshape((1, -1))

    n_traj = ic.shape[0]

    if tg_ic is None:
        tg_ic = np.eye(ic.shape[1])

    tg_ic_sav = tg_ic.copy()

    if len(tg_ic.shape) == 1:
        tg_ic = tg_ic.reshape((1, -1, 1))
        ict = tg_ic.copy()
        for i in range(n_traj-1):
            ict = np.concatenate((ict, tg_ic))
        tg_ic = ict
    elif len(tg_ic.shape) == 2:
        if tg_ic.shape[0] == n_traj:
            tg_ic = tg_ic[..., np.newaxis]
        else:
            tg_ic = tg_ic[np.newaxis, ...]
            tg_ic = np.swapaxes(tg_ic, 1, 2)
            ict = tg_ic.copy()
            for i in range(n_traj-1):
                ict = np.concatenate((ict, tg_ic))
            tg_ic = ict
    elif len(tg_ic.shape) == 3:
        if tg_ic.shape[1] != ic.shape[1]:
            tg_ic = np.swapaxes(tg_ic, 1, 2)

    # Default is RK4
    if a is None and b is None and c is None:
        c = np.array([0., 0.5, 0.5, 1.])
        b = np.array([1./6, 1./3, 1./3, 1./6])
        a = np.zeros((len(c), len(b)))
        a[1, 0] = 0.5
        a[2, 1] = 0.5
        a[3, 2] = 1.

    if forward:
        time_direction = 1
    else:
        time_direction = -1

    time = np.concatenate((np.arange(t0, t, dt), np.full((1,), t)))

    if boundary is None:
        boundary = _zeros_func

    inv = 1.
    if inverse:
        inv *= -1.

    recorded_traj, recorded_fmatrix = _integrate_runge_kutta_tgls_jit(f, fjac, time, ic, tg_ic,
                                                                      time_direction, write_steps,
                                                                      b, c, a, adjoint, inv, boundary)

    if len(tg_ic_sav.shape) == 2:
        if recorded_fmatrix.shape[1:3] != tg_ic_sav.shape:
            recorded_fmatrix = np.swapaxes(recorded_fmatrix, 1, 2)

    elif len(tg_ic_sav.shape) == 3:
        if tg_ic_sav.shape[1] != ic.shape[1]:
            if recorded_fmatrix.shape[:3] != tg_ic_sav.shape:
                recorded_fmatrix = np.swapaxes(recorded_fmatrix, 1, 2)

    if write_steps > 0:
        if forward:
            if time[::write_steps][-1] == time[-1]:
                return time[::write_steps], np.squeeze(recorded_traj), np.squeeze(recorded_fmatrix)
            else:
                return np.concatenate((time[::write_steps], np.full((1,), t))), np.squeeze(recorded_traj),\
                       np.squeeze(recorded_fmatrix)
        else:
            rtime = reverse(time[::-write_steps])
            if rtime[0] == time[0]:
                return rtime, np.squeeze(recorded_traj), np.squeeze(recorded_fmatrix)
            else:
                return np.concatenate((np.full((1,), t0), rtime)), np.squeeze(recorded_traj),\
                       np.squeeze(recorded_fmatrix)

    else:
        return time[-1], np.squeeze(recorded_traj), np.squeeze(recorded_fmatrix)


@njit
def _integrate_runge_kutta_tgls_jit(f, fjac, time, ic, tg_ic, time_direction, write_steps, b, c, a,
                                    adjoint, inverse, boundary):

    n_traj = ic.shape[0]
    n_dim = ic.shape[1]

    s = len(b)

    if write_steps == 0:
        n_records = 1
    else:
        tot = time[::write_steps]
        n_records = len(tot)
        if tot[-1] != time[-1]:
            n_records += 1
    recorded_traj = np.zeros((n_traj, n_dim, n_records))
    recorded_fmatrix = np.zeros((n_traj, tg_ic.shape[1], tg_ic.shape[2], n_records))
    if time_direction == -1:
        directed_time = reverse(time)
    else:
        directed_time = time

    for i_traj in range(n_traj):
        y = ic[i_traj].copy()
        fm = tg_ic[i_traj].copy()
        recorded_traj[i_traj, :, 0] = ic[i_traj]
        recorded_fmatrix[i_traj, :, :, 0] = tg_ic[i_traj]
        k = np.zeros((s, n_dim))
        km = np.zeros((s, tg_ic.shape[1], tg_ic.shape[2]))
        iw = 0
        for ti, (tt, dt) in enumerate(zip(directed_time[:-1], np.diff(directed_time))):

            if write_steps > 0 and np.mod(ti, write_steps) == 0:
                recorded_traj[i_traj, :, iw] = y
                recorded_fmatrix[i_traj, :, :, iw] = fm
                iw += 1

            k.fill(0.)
            km.fill(0.)
            for i in range(s):
                y_s = y + dt * a[i] @ k
                k[i] = f(tt + c[i] * dt, y_s)
                km_s = fm.copy()
                for j in range(len(a[i])):
                    km_s += dt * a[i, j] * km[j]
                hom = inverse * _tangent_linear_system(fjac, tt + c[i] * dt, y_s, km_s, adjoint)
                inhom = boundary(tt + c[i] * dt, y_s)
                km[i] = (hom.T + inhom.T).T
            y_new = y + dt * b @ k
            fm_new = fm.copy()
            for j in range(len(b)):
                fm_new += dt * b[j] * km[j]
            y = y_new
            fm = fm_new

        recorded_traj[i_traj, :, -1] = y
        recorded_fmatrix[i_traj, :, :, -1] = fm

    return recorded_traj[:, :, ::time_direction], recorded_fmatrix[:, :, :, ::time_direction]


if __name__ == "__main__":
    import matplotlib.pyplot as plt
    from scipy.integrate import odeint

    @njit
    def f(t, x):
        return - np.array([1., 2., 3.]) * x

    def fr(x, t):
        return f(t, x)

    ic = np.random.randn(6).reshape(2, 3)

    time, r = integrate_runge_kutta(f, 0., 10., 0.01, ic=ic, write_steps=3)

    t = np.arange(0., 10., 0.01)
    t = np.concatenate((t[::3], np.full((1,), 10.)))
    rl = list()
    for i in range(ic.shape[0]):
        rl.append(odeint(fr, ic[i], t).T)

    plt.figure()
    for i in range(ic.shape[0]):
        p, = plt.plot(time, r[i, 0])
        c = p.get_color()
        plt.plot(t, rl[i][0], color=c, ls='--')
        for j in range(1, ic.shape[1]):
            p, = plt.plot(time, r[i, j], color=c)
            plt.plot(t, rl[i][j], color=c, ls='--')

    timet, rt = integrate_runge_kutta(f, 0., 10., 0.01, ic=ic, forward=False, write_steps=3)
    rlt = list()
    for i in range(ic.shape[0]):
        rlt.append(odeint(fr, ic[i], reverse(t)).T)

    plt.figure()
    for i in range(ic.shape[0]):
        p, = plt.plot(timet, rt[i, 0])
        c = p.get_color()
        plt.plot(t, reverse(rlt[i][0]), color=c, ls='--')
        for j in range(1, ic.shape[1]):
            p, = plt.plot(timet, rt[i, j], color=c)
            plt.plot(t, reverse(rlt[i][j]), color=c, ls='--')

    tt, re = integrate_runge_kutta(f, 0., 10., 0.01, ic=ic, write_steps=0)
    print(tt)
    print(r[0, :, -1], re[0])
    plt.show(block=False)

    a = 0.25
    F = 16.
    G = 3.
    b = 6.


    @njit
    def DfL84(t, x):
        return np.array([[     -a        , -2. * x[1], -2. * x[2]],
                         [x[1] - b * x[2], -1. + x[0], -b * x[0]],
                         [b * x[1] + x[2],  b * x[0], -1. + x[0]]])

    @njit
    def fL84(t, x):
        xx = -x[1] ** 2 - x[2] ** 2 - a * x[0] + a * F
        yy = x[0] * x[1] - b * x[0] * x[2] - x[1] + G
        zz = b * x[0] * x[1] + x[0] * x[2] - x[2]
        return np.array([xx, yy, zz])

    def fL84r(x, t):
        return fL84(t, x)

    tt, ic_L84 = integrate_runge_kutta(fL84, 0., 10000., 0.01, write_steps=0)

    print(ic_L84)

    ttt1, irkt1, fm_irkt1 = integrate_runge_kutta_tgls(fL84, DfL84, 0., 0.01, 0.001, ic=ic_L84, write_steps=1)
    bttt1, birkt1, bfm_irkt1 = integrate_runge_kutta_tgls(fL84, DfL84, 0., 0.01, 0.001, ic=irkt1[:, -1],
                                                          forward=False, write_steps=1, adjoint=True, inverse=True)

    plt.figure()
    for i in range(len(ic_L84)):
        for j in range(len(ic_L84)):
            p, = plt.plot(ttt1, fm_irkt1[i, j])
            c = p.get_color()
            plt.plot(bttt1, bfm_irkt1[i, j], ls='--', color=c)

    plt.figure()
    for i in range(len(ic_L84)):
        p, = plt.plot(ttt1, irkt1[i])
        c = p.get_color()
        plt.plot(bttt1, birkt1[i], ls='--', color=c)

    vec = np.random.randn(3)
    vec = vec/np.linalg.norm(vec)
    for i in range(1,13):
        lam = 2.**(-i)

        dy = lam * vec

        tx, dx = integrate_runge_kutta(fL84, 0., 0.001, 0.001, ic=ic_L84, write_steps=0)
        txp, dxp = integrate_runge_kutta(fL84, 0., 0.001, 0.001, ic=ic_L84+dy, write_steps=0)

        dy1 = dxp - dx

        txtl, dxtl, dytl = integrate_runge_kutta_tgls(fL84, DfL84, 0., 0.001, 0.001, ic=ic_L84, tg_ic=dy, write_steps=0)

        print('Perturbation size: ', dy @ dy)
        print('Resulting difference in trajectory: (eps ~ 2^-'+str(i)+')')
        print('diff: ', dy1 @ dy1)
        print('tl: ', dytl @ dytl)
        print('ratio: ', (dy1 @ dy1) / (dytl @ dytl))



"""
    Symbolic Inner products module
    ==============================

    Inner products between the truncated set of basis functions :math:`\\phi_i` for the ocean and land fields, and
    :math:`F_i` for the atmosphere fields (see :ref:`files/model/oro_model:Projecting the equations on a set of basis functions`).

    Notes
    -----

    These inner products are computed symbolically using `Sympy`_, and thus support arbitrary (but symbolic) basis functions.

    Description of the classes
    --------------------------

    The three classes computing and holding the inner products of the basis functions are:

    * :class:`AtmosphericSymbolicInnerProducts`
    * :class:`OceanicSymbolicInnerProducts`
    * :class:`GroundSymbolicInnerProducts`

    .. _Sympy: https://www.sympy.org/
"""

import sparse as sp
from pebble import ProcessPool as Pool
from concurrent.futures import TimeoutError
from multiprocessing import cpu_count
from sympy.utilities.iterables import multiset_permutations

from qgs.params.params import QgParams
from qgs.inner_products.base import AtmosphericInnerProducts, OceanicInnerProducts, GroundInnerProducts
from qgs.inner_products.definition import StandardSymbolicInnerProductDefinition
from scipy.integrate import dblquad
from sympy import ImmutableSparseMatrix, ImmutableSparseNDimArray, lambdify

# TODO: - Add warnings if trying to connect analytic and symbolic inner products together


class AtmosphericSymbolicInnerProducts(AtmosphericInnerProducts):
    """Class which contains all the atmospheric inner products coefficients needed for the tendencies
    tensor :class:`~.tensors.qgtensor.QgsTensor` computation, computed with analytic formula.

    Parameters
    ----------
    params: QgParams or list
        An instance of model's parameters object or a list in the form [aspect_ratio, atmospheric_basis, basis, oog, oro_basis].
        If a list is provided, `aspect_ratio` is the aspect ratio of the domain, `atmospheric_basis` is a SymbolicBasis with
        the modes of the atmosphere, and `ocean_basis` is either `None` or a SymbolicBasis object with the modes of
        the ocean or the ground. Finally, `oog` indicates if it is an ocean or a ground component that is connected,
        by setting it to `ocean` or to 'ground', and in this latter case, `oro_basis` indicates on which basis the orography is developed.
    stored: bool, optional
        Indicate if the inner product must be stored or computed on the fly. Default to `True`
    inner_product_definition: None or InnerProductDefinition, optional
        The definition of the inner product being used. If `None`, use the canonical StandardInnerProductDefinition object.
        Default to `None`.
    interaction_inner_product_definition: None or InnerProductDefinition, optional
        The definition of the inner product being used for the interaction with the other components, i.e. to compute the inner products with the other component base of funcitons.
        If `None`, use the `inner_product_definition` provided.
        Default to `None`.
    num_threads: int or None, optional
        Number of threads to use to compute the symbolic inner products. If `None` use all the cpus available.
        Default to `None`.
    quadrature: bool, optional
        If `True', compute the inner products with a quadrature instead of the symbolic integration.
        If `True` Disable the `timeout` parameter.
        Default to `True`.
    timeout: int or float or bool or None, optional
        The timeout for the computation of each inner product. After the timeout, compute the inner product with a quadrature instead of symbolic integration.
        If `None` or `False`, no timeout occurs.
        Default to `None`.
    dynTinnerproducts: bool, optional
        If the inner products are stored, allow to compute or not the inner products corresponding to the dynamic temperature tendencies.
        Supersedes the parameters in params if provided.
        Default to `False`.
    T4innerproducts: bool, optional
        If the inner products are stored, allow to compute or not the inner products corresponding to the :math:`T^4` tendencies.
        Compute the inner products corresponding to the dynamic temperature tendencies as well.
        Supersedes the parameters in params if provided.
        Default to `False`.

    Attributes
    ----------
    n: float
        The aspect ratio of the domain.
    atmospheric_basis: SymbolicBasis
        Object holding the symbolic modes of the atmosphere.
    oceanic_basis: None or SymbolicBasis
        Object holding the symbolic modes of the ocean (or `None` if there is no ocean).
    connected_to_ocean: bool
        Indicate if the atmosphere is connected to an ocean.
    stored: bool
        Indicate if the inner product must be stored or computed on the fly.
    ip: InnerProductDefinition
        Object defining the inner product.
    iip: InnerProductDefinition
        Object defining the interaction inner product.
    subs: list(tuple)
        List of 2-tuples containing the substitutions to be made with the functions after the inner products
        symbolic computation.
    """

    def __init__(self, params=None, stored=True, inner_product_definition=None, interaction_inner_product_definition=None,
                 num_threads=None, quadrature=True, timeout=None, dynTinnerproducts=None, T4innerproducts=None, return_symbolic=False, make_substitution=True):

        AtmosphericInnerProducts.__init__(self)

        self.quadrature = quadrature
        if quadrature:
            timeout = True

        if params is not None:
            if isinstance(params, QgParams):
                self.n = params.scale_params.n
                self.atmospheric_basis = params.atmospheric_basis
                if params.oceanic_basis is not None:
                    goc_basis = params.oceanic_basis
                    oog = "ocean"
                elif params.ground_basis is not None:
                    goc_basis = params.ground_basis
                    oog = "ground"
                    oro_basis = params.ground_params.orographic_basis
                else:
                    goc_basis = None
                    oog = ""
                    oro_basis = params.ground_params.orographic_basis
                if T4innerproducts is not None:
                    self._T4 = T4innerproducts
                else:
                    self._T4 = params.T4
                if dynTinnerproducts is not None:
                    self._dynamic_T = dynTinnerproducts
                else:
                    self._dynamic_T = params.dynamic_T
            else:
                self.n = params[0]
                self.atmospheric_basis = params[1]
                goc_basis = params[2]
                oog = params[3]
                oro_basis = params[4]
                if T4innerproducts is not None:
                    self._T4 = T4innerproducts
                else:
                    self._T4 = False
                if dynTinnerproducts is not None:
                    self._dynamic_T = dynTinnerproducts
                else:
                    self._dynamic_T = False
            self._gh = None
        else:  # initialize an empty inner product object
            self.n = None
            self.atmospheric_basis = None
            goc_basis = None
            oog = ""
            self._gh = None
            stored = False
            oro_basis = ""
            if T4innerproducts is not None:
                self._T4 = T4innerproducts
            else:
                self._T4 = False
            if dynTinnerproducts is not None:
                self._dynamic_T = dynTinnerproducts
            else:
                self._dynamic_T = False

        self.oceanic_basis = None
        self.connected_to_ocean = False

        self.ground_basis = None
        self.connected_to_ground = False

        self.return_symbolic = return_symbolic
        if return_symbolic:
            self.mk_subs = make_substitution
            if self.mk_subs:
                self.subs = [(self.n.symbol, self.n)]
            else:
                self.subs = None
        else:
            self.mk_subs = True
            self.subs = [(self.n.symbol, self.n)]

        if inner_product_definition is None:
            self.ip = StandardSymbolicInnerProductDefinition()
        else:
            self.ip = inner_product_definition

        if interaction_inner_product_definition is None:
            self.iip = self.ip
        else:
            self.iip = interaction_inner_product_definition

        self.stored = stored
        if stored:
            self.compute_inner_products(num_threads, timeout)

        if goc_basis is not None:
            if oog == 'ocean':
                self.connect_to_ocean(goc_basis, num_threads, timeout)
            else:
                self.connect_to_ground(goc_basis, oro_basis, num_threads, timeout)

    def _F(self, i):
        if self.atmospheric_basis is not None:
            return self.atmospheric_basis.functions[i]

    def _phi(self, i):
        if self.oceanic_basis is not None:
            return self.oceanic_basis.functions[i]
        elif self.ground_basis is not None:
            return self.ground_basis.functions[i]

    def connect_to_ocean(self, ocean_basis, num_threads=None, timeout=None):
        """Connect the atmosphere to an ocean.

        Parameters
        ----------
        ocean_basis: SymbolicBasis or OceanicSymbolicInnerProducts
            Basis of function of the ocean or a symbolic oceanic inner products object containing the basis.
        num_threads: int or None, optional
            Number of threads to use to compute the symbolic inner products. If `None` use all the cpus available.
            Default to `None`.
        timeout: int or float or bool or None, optional
            The timeout for the computation of each inner product. After the timeout, compute the inner product with a quadrature instead of symbolic integration.
            If `True`, force the timeout and compute directly the inner product with a quadrature instead of trying to do the integration symbolically.
            If `None` or `False`, no timeout occurs.
            Default to `None`.
       """
        if isinstance(ocean_basis, OceanicSymbolicInnerProducts):
            ocean_basis = ocean_basis.oceanic_basis
        self.ground_basis = None
        self.connected_to_ground = False

        self.oceanic_basis = ocean_basis
        self.connected_to_ocean = True

        if self.stored:

            if num_threads is None:
                num_threads = cpu_count()

            with Pool(max_workers=num_threads) as pool:
                if self.mk_subs:
                    subs = self.subs + self.atmospheric_basis.substitutions + self.oceanic_basis.substitutions
                else:
                    subs = self.subs

                noc = len(ocean_basis)
                if self.return_symbolic:
                    self._gh = None
                    self._d = None
                    self._s = None
                    self._v = None
                else:
                    self._gh = None
                    self._d = sp.zeros((self.natm, noc), dtype=float, format='dok')
                    self._s = sp.zeros((self.natm, noc), dtype=float, format='dok')
                    if self._T4 or self._dynamic_T:
                        self._v = sp.zeros((self.natm, noc, noc, noc, noc), dtype=float, format='dok')

                # d inner products
                args_list = [[(i, j), self.iip.ip_lap, (self._F(i), self._phi(j))] for i in range(self.natm)
                             for j in range(noc)]

                output = _parallel_compute(pool, args_list, subs, self._d, timeout, symbolic_int=not self.mk_subs)
                if self.return_symbolic:
                    self._d = ImmutableSparseMatrix(self.natm, noc, output)
                else:
                    self._d = self._d.to_coo()

                # s inner products
                args_list = [[(i, j), self.iip.symbolic_inner_product, (self._F(i), self._phi(j))] for i in range(self.natm)
                             for j in range(noc)]

                output = _parallel_compute(pool, args_list, subs, self._s, timeout, symbolic_int=not self.mk_subs)
                if self.return_symbolic:
                    self._s = ImmutableSparseMatrix(self.natm, noc, output)
                else:
                    self._s = self._s.to_coo()

                if self._T4:
                    # v inner products
                    args_list = [[(i, j, k, ell, m), self.ip.symbolic_inner_product, (self._F(i), self._phi(j) * self._phi(k) * self._phi(ell) * self._phi(m))] for i in range(self.natm)
                                 for j in range(noc) for k in range(j, noc) for ell in range(k, noc) for m in range(ell, noc)]

                    output = _parallel_compute(pool, args_list, subs, self._v, timeout, permute=True, symbolic_int=not self.mk_subs)
                elif self._dynamic_T:
                    # v inner products
                    args_list = [[(i, 0, 0, 0, m), self.ip.symbolic_inner_product, (self._F(i), self._phi(0) * self._phi(0) * self._phi(0) * self._phi(m))]
                                 for i in range(self.natm) for m in range(noc)]

                    output = _parallel_compute(pool, args_list, subs, self._v, timeout, permute=True, symbolic_int=not self.mk_subs)

                if self._T4 or self._dynamic_T:
                    if self.return_symbolic:
                        self._v = ImmutableSparseNDimArray(output, shape=(self.natm, noc, noc, noc, noc))
                    else:
                        self._v = self._v.to_coo()            

    def connect_to_ground(self, ground_basis, orographic_basis, num_threads=None, timeout=None):
        """Connect the atmosphere to the ground.

        Parameters
        ----------
        ground_basis: SymbolicBasis or GroundSymbolicInnerProducts
            Basis of function of the ground or a symbolic ground inner products object containing the basis.
        orographic_basis: str
            String to select which component basis modes to use to develop the orography in series.
            Can be either 'atmospheric' or 'ground'.
        num_threads: int or None, optional
            Number of threads to use to compute the symbolic inner products. If `None` use all the cpus available.
            Default to `None`.
        timeout: int or float or bool or None, optional
            The timeout for the computation of each inner product. After the timeout, compute the inner product with a quadrature instead of symbolic integration.
            If `True`, force the timeout and compute directly the inner product with a quadrature instead of trying to do the integration symbolically.
            If `None` or `False`, no timeout occurs.
            Default to `None`.
        """

        if isinstance(ground_basis, GroundSymbolicInnerProducts):
            ground_basis = ground_basis.ground_basis

        self.oceanic_basis = None
        self.connected_to_ocean = False

        self.ground_basis = ground_basis
        self.connected_to_ground = True

        if self.stored:
            if num_threads is None:
                num_threads = cpu_count()

            with Pool(max_workers=num_threads) as pool:
                if self.mk_subs:
                    subs = self.subs + self.atmospheric_basis.substitutions + self.ground_basis.substitutions
                else:
                    subs = self.subs

                ngr = len(ground_basis)
                if self.return_symbolic:
                    self._gh = None
                    self._d = None
                    self._s = None
                    self._v = None
                else:
                    if orographic_basis == "atmospheric":
                        self._gh = None
                    else:
                        self._gh = sp.zeros((self.natm, self.natm, ngr), dtype=float, format='dok')
                    self._d = None
                    self._s = sp.zeros((self.natm, ngr), dtype=float, format='dok')
                    if self._T4 or self._dynamic_T:
                        self._v = sp.zeros((self.natm, ngr, ngr, ngr, ngr), dtype=float, format='dok')

                # s inner products
                args_list = [[(i, j), self.iip.symbolic_inner_product, (self._F(i), self._phi(j))] for i in range(self.natm)
                             for j in range(ngr)]

                output = _parallel_compute(pool, args_list, subs, self._s, timeout, symbolic_int=not self.mk_subs)
                if self.return_symbolic:
                    self._s = ImmutableSparseMatrix(self.natm, ngr, output)
                else:
                    self._s = self._s.to_coo()

                # gh inner products
                if orographic_basis != "atmospheric":
                    args_list = [[(i, j, k), self.iip.ip_jac, (self._F(i), self._F(j), self._phi(k))] for i in range(self.natm)
                                 for j in range(self.natm) for k in range(ngr)]

                    output = _parallel_compute(pool, args_list, subs, self._gh, timeout, symbolic_int=not self.mk_subs)
                    if self.return_symbolic:
                        self._gh = ImmutableSparseNDimArray(output, shape=(self.natm, self.natm, ngr))
                    else:
                        if self._gh is not None:
                            self._gh = self._gh.to_coo()

                if self._T4:
                    # v inner products
                    args_list = [[(i, j, k, ell, m), self.ip.symbolic_inner_product, (self._F(i), self._phi(j) * self._phi(k) * self._phi(ell) * self._phi(m))] for i in range(self.natm)
                                 for j in range(ngr) for k in range(j, ngr) for ell in range(k, ngr) for m in range(ell, ngr)]

                    output = _parallel_compute(pool, args_list, subs, self._v, timeout, permute=True, symbolic_int=not self.mk_subs)
                elif self._dynamic_T:
                    # v inner products
                    args_list = [[(i, 0, 0, 0, m), self.ip.symbolic_inner_product, (self._F(i), self._phi(0) * self._phi(0) * self._phi(0) * self._phi(m))]
                                 for i in range(self.natm) for m in range(ngr)]

                    output = _parallel_compute(pool, args_list, subs, self._v, timeout, permute=True, symbolic_int=not self.mk_subs)

                if self._T4 or self._dynamic_T:
                    if self.return_symbolic:
                        self._v = ImmutableSparseNDimArray(output, shape=(self.natm, ngr, ngr, ngr, ngr))
                    else:
                        self._v = self._v.to_coo()

    def compute_inner_products(self, num_threads=None, timeout=None):
        """Function computing and storing all the inner products at once.

        Parameters
        ----------
        num_threads: int or None, optional
            Number of threads to use to compute the symbolic inner products. If `None` use all the cpus available.
            Default to `None`.
        timeout: int or float or bool or None, optional
            The timeout for the computation of each inner product. After the timeout, compute the inner product with a quadrature instead of symbolic integration.
            If `True`, force the timeout and compute directly the inner product with a quadrature instead of trying to do the integration symbolically.
            If `None` or `False`, no timeout occurs.
            Default to `None`.
        """
        if self.return_symbolic:
            self._a = None
            self._u = None
            self._c = None
            self._b = None
            self._g = None
            self._z = None
        else:
            self._a = sp.zeros((self.natm, self.natm), dtype=float, format='dok')
            self._u = sp.zeros((self.natm, self.natm), dtype=float, format='dok')
            self._c = sp.zeros((self.natm, self.natm), dtype=float, format='dok')
            self._b = sp.zeros((self.natm, self.natm, self.natm), dtype=float, format='dok')
            self._g = sp.zeros((self.natm, self.natm, self.natm), dtype=float, format='dok')
            if self._T4 or self._dynamic_T:
                self._z = sp.zeros((self.natm, self.natm, self.natm, self.natm, self.natm), dtype=float, format='dok')

        if self.stored:
            if num_threads is None:
                num_threads = cpu_count()

            with Pool(max_workers=num_threads) as pool:
                if self.mk_subs:
                    subs = self.subs + self.atmospheric_basis.substitutions
                else:
                    subs = self.subs

                # a inner products
                args_list = [[(i, j), self.ip.ip_lap, (self._F(i), self._F(j))] for i in range(self.natm)
                             for j in range(self.natm)]
                
                output = _parallel_compute(pool, args_list, subs, self._a, timeout, symbolic_int=not self.mk_subs)
                if self.return_symbolic:
                    self._a = ImmutableSparseMatrix(self.natm, self.natm, output)
                else:
                    self._a = self._a.to_coo()

                # u inner products
                args_list = [[(i, j), self.ip.symbolic_inner_product, (self._F(i), self._F(j))] for i in range(self.natm)
                             for j in range(self.natm)]

                output = _parallel_compute(pool, args_list, subs, self._u, timeout, symbolic_int=not self.mk_subs)
                if self.return_symbolic:
                    self._u = ImmutableSparseMatrix(self.natm, self.natm, output)
                else:
                    self._u = self._u.to_coo()

                # c inner products
                args_list = [[(i, j), self.ip.ip_diff_x, (self._F(i), self._F(j))] for i in range(self.natm)
                             for j in range(self.natm)]

                output = _parallel_compute(pool, args_list, subs, self._c, timeout, symbolic_int=not self.mk_subs)
                if self.return_symbolic:
                    self._c = ImmutableSparseMatrix(self.natm, self.natm, output)
                else:
                    self._c = self._c.to_coo()

                # b inner products
                args_list = [[(i, j, k), self.ip.ip_jac_lap, (self._F(i), self._F(j), self._F(k))] for i in range(self.natm)
                             for j in range(self.natm) for k in range(self.natm)]

                output = _parallel_compute(pool, args_list, subs, self._b, timeout, symbolic_int=not self.mk_subs)
                if self.return_symbolic:
                    self._b = ImmutableSparseNDimArray(output, shape=(self.natm, self.natm, self.natm))
                else:
                    self._b = self._b.to_coo()

                # g inner products
                args_list = [[(i, j, k), self.ip.ip_jac, (self._F(i), self._F(j), self._F(k))] for i in range(self.natm)
                             for j in range(self.natm) for k in range(self.natm)]

                output = _parallel_compute(pool, args_list, subs, self._g, timeout, symbolic_int=not self.mk_subs)
                if self.return_symbolic:
                    self._g = ImmutableSparseNDimArray(output, shape=(self.natm, self.natm, self.natm))
                else:
                    self._g = self._g.to_coo()

                if self._T4:
                    # z inner products
                    args_list = [[(i, j, k, ell, m), self.ip.symbolic_inner_product, (self._F(i), self._F(j) * self._F(k) * self._F(ell) * self._F(m))] for i in range(self.natm)
                                 for j in range(self.natm) for k in range(j, self.natm) for ell in range(k, self.natm) for m in range(ell, self.natm)]

                    output = _parallel_compute(pool, args_list, subs, self._z, timeout, permute=True, symbolic_int=not self.mk_subs)
                elif self._dynamic_T:
                    # z inner products
                    args_list = [[(i, 0, 0, 0, m), self.ip.symbolic_inner_product, (self._F(i), self._F(0) * self._F(0) * self._F(0) * self._F(m))]
                                 for i in range(self.natm) for m in range(self.natm)]

                    output = _parallel_compute(pool, args_list, subs, self._z, timeout, permute=True, symbolic_int=not self.mk_subs)

                if self._T4 or self._dynamic_T:
                    if self.return_symbolic:
                        self._z = ImmutableSparseNDimArray(output, shape=(self.natm, self.natm, self.natm, self.natm, self.natm))
                    else:
                        self._z = self._z.to_coo()

    @property
    def natm(self):
        """Number of atmospheric modes."""
        return len(self.atmospheric_basis.functions)

    # !-----------------------------------------------------!
    # ! Inner products in the equations for the atmosphere  !
    # !-----------------------------------------------------!

    def _integrate(self, subs, args):
        if not self.mk_subs:
            res = _apply(args)
            return res[1]
        
        if self.quadrature:
            res = _num_apply(args)
            return res[1]
        else:
            res = _apply(args)[1]
            if subs is not None:
                return float(res.subs(subs))
            else:
                return res

    def a(self, i, j):
        """Function to compute the matrix of the eigenvalues of the Laplacian (atmospheric): :math:`a_{i, j} = (F_i, {\\nabla}^2 F_j)`."""
        if not self.stored:
            if self.mk_subs:
                subs = self.subs + self.atmospheric_basis.substitutions
            else:
                subs = self.subs

            args = ((i, j), self.ip.ip_lap, (self._F(i), self._F(j)), subs)
            return self._integrate(subs, args)
        else:
            return self._a[i, j]

    def u(self, i, j):
        """Function to compute the matrix of inner product: :math:`u_{i, j} = (F_i, F_j)`."""
        if not self.stored:
            if self.mk_subs:
                subs = self.subs + self.atmospheric_basis.substitutions
            else:
                subs = self.subs

            args = ((i, j), self.ip.symbolic_inner_product, (self._F(i), self._F(j)), subs)
            return self._integrate(subs, args)
        else:
            return self._u[i, j]

    def b(self, i, j, k):
        """Function to compute the tensors holding the Jacobian inner products: :math:`b_{i, j, k} = (F_i, J(F_j, \\nabla^2 F_k))`."""
        if not self.stored:
            if self.mk_subs:
                subs = self.subs + self.atmospheric_basis.substitutions
            else:
                subs = self.subs
            
            args = ((i, j, k), self.ip.ip_jac_lap, (self._F(i), self._F(j), self._F(k)), subs)
            return self._integrate(subs, args)
        else:
            return self._b[i, j, k]

    def c(self, i, j):
        """Function to compute the matrix of beta terms for the atmosphere: :math:`c_{i,j} = (F_i, \\partial_x F_j)`."""
        if not self.stored:
            if self.mk_subs:
                subs = self.subs + self.atmospheric_basis.substitutions
            else:
                subs = self.subs

            args = ((i, j), self.ip.ip_diff_x, (self._F(i), self._F(j)), subs)
            return self._integrate(subs, args)
        else:
            return self._c[i, j]

    def g(self, i, j, k):
        """Function to compute tensors holding the Jacobian inner products: :math:`g_{i,j,k} = (F_i, J(F_j, F_k))`."""
        if not self.stored:
            if self.mk_subs:
                subs = self.subs + self.atmospheric_basis.substitutions
            else:
                subs = self.subs

            args = ((i, j, k), self.ip.ip_jac, (self._F(i), self._F(j), self._F(k)), subs)
            return self._integrate(subs, args)
        else:
            return self._g[i, j, k]

    def gh(self, i, j, k):
        """Function to compute tensors holding the Jacobian inner products: :math:`g_{i,j,k} = (F_i, J(F_j, \\phi_k))`."""

        if self.oceanic_basis or self.connected_to_ground:
            if self.stored and self._gh is not None:
                return self._gh[i, j, k]
            else:
                if self.mk_subs:
                    if self.connected_to_ocean:
                        extra_subs = self.oceanic_basis.substitutions
                    elif self.connected_to_ground:
                        extra_subs = self.ground_basis.substitutions
                    else:
                        extra_subs = None
                    subs = self.subs + self.atmospheric_basis.substitutions + extra_subs
                else:
                    subs = self.subs

                args = ((i, j, k), self.iip.ip_jac, (self._F(i), self._F(j), self._phi(k)), subs)
                return self._integrate(subs, args)
        else:
            return 0

    def s(self, i, j):
        """Function to compute the forcing (thermal) of the ocean on the atmosphere: :math:`s_{i,j} = (F_i, \\phi_j)`."""
        if self.connected_to_ocean or self.connected_to_ground:
            if self.stored and self._s is not None:
                return self._s[i, j]
            else:
                if self.mk_subs:
                    if self.connected_to_ocean:
                        extra_subs = self.oceanic_basis.substitutions
                    elif self.connected_to_ground:
                        extra_subs = self.ground_basis.substitutions
                    else:
                        extra_subs = None
                    subs = self.subs + self.atmospheric_basis.substitutions + extra_subs
                else:
                    subs = self.subs

                args = ((i, j), self.iip.symbolic_inner_product, (self._F(i), self._phi(j)), subs)
                return self._integrate(subs, args)
        else:
            return 0

    def d(self, i, j):
        """Function to compute the forcing of the ocean on the atmosphere: :math:`d_{i,j} = (F_i, \\nabla^2 \\phi_j)`."""
        if self.connected_to_ocean or self.connected_to_ground:
            if self.stored and self._d is not None:
                return self._d[i, j]
            else:
                if self.mk_subs:
                    if self.connected_to_ocean:
                        extra_subs = self.oceanic_basis.substitutions
                    elif self.connected_to_ground:
                        extra_subs = self.ground_basis.substitutions
                    else:
                        extra_subs = None

                    subs = self.subs + self.atmospheric_basis.substitutions + extra_subs
                else:
                    subs = self.subs

                args = ((i, j), self.iip.ip_lap, (self._F(i), self._phi(j)), subs)
                return self._integrate(subs, args)
        else:
            return 0

    def z(self, i, j, k, l, m):
        """Function to compute the :math:`T^4` temperature forcing for the radiation lost by atmosphere to space & ground/ocean: :math:`z_{i,j,k,l,m} = (F_i, F_j F_k F_l F_m)`."""
        if not self.stored:
            if self.mk_subs:
                subs = self.subs + self.atmospheric_basis.substitutions
            else:
                subs = self.subs

            args = ((i, j, k, l, m), self.ip.symbolic_inner_product, (self._F(i), self._F(j) * self._F(k) * self._F(l) * self._F(m)), subs)
            if self.quadrature:
                res = _num_apply(args)
                return res[1]
            else:
                res = _apply(args)[1]
                return float(res.subs(subs))
        else:
            return self._z[i, j, k, l, m]

    def v(self, i, j, k, l, m):
        """Function to compute the :math:`T^4` temperature forcing of the ocean on the atmosphere: :math:`v_{i,j,k,l,m} = (F_i, \\phi_j \\phi_k \\phi_l \\phi_m)`."""
        if not self.stored:
            if self.mk_subs:
                subs = self.subs + self.atmospheric_basis.substitutions
            else:
                subs = self.subs

            args = ((i, j, k, l, m), self.ip.symbolic_inner_product, (self._F(i), self._phi(j) * self._phi(k) * self._phi(l) * self._phi(m)), subs)
            if self.quadrature:
                res = _num_apply(args)
                return res[1]
            else:
                res = _apply(args)[1]
                return float(res.subs(subs))
        else:
            return self._v[i, j, k, l, m]


class OceanicSymbolicInnerProducts(OceanicInnerProducts):
    """Class which contains all the oceanic inner products coefficients needed for the tendencies
    tensor :class:`~.tensors.qgtensor.QgsTensor` computation.

    Parameters
    ----------
    params: QgParams or list
        An instance of model's parameters object or a list in the form [aspect_ratio, ocean_basis, atmospheric_basis].
        If a list is provided, `aspect_ratio` is the aspect ratio of the domain, `ocean_basis` is a SymbolicBasis object
        with the modes of the ocean, and `atmospheric_basis` is either a SymbolicBasis with the modes of the atmosphere
        or `None` if there is no atmosphere.
    stored: bool, optional
        Indicate if the inner product must be stored or computed on the fly. Default to `True`
    inner_product_definition: None or InnerProductDefinition, optional
        The definition of the inner product being used. If `None`, use the canonical StandardInnerProductDefinition object.
        Default to `None`.
    interaction_inner_product_definition: None or InnerProductDefinition, optional
        The definition of the inner product being used for the interaction with the other components, i.e. to compute the inner products with the other component base of funcitons.
        If `None`, use the `inner_product_definition` provided.
        Default to `None`.
    num_threads: int or None, optional
        Number of threads to use to compute the symbolic inner products. If `None` use all the cpus available.
        Default to `None`.
    quadrature: bool, optional
        If `True', compute the inner products with a quadrature instead of the symbolic integration.
        If `True` Disable the `timeout` parameter.
        Default to `True`.
    timeout: int or float or bool or None, optional
        The timeout for the computation of each inner product. After the timeout, compute the inner product with a quadrature instead of symbolic integration.
        If `None` or `False`, no timeout occurs.
        Default to `None`.
    dynTinnerproducts: bool, optional
        If the inner products are stored, allow to compute or not the inner products corresponding to the dynamic temperature tendencies.
        Supersedes the parameters in params if provided.
        Default to `False`.
    T4innerproducts: bool, optional
        If the inner products are stored, allow to compute or not the inner products corresponding to the :math:`T^4` tendencies.
        Compute the inner products corresponding to the dynamic temperature tendencies as well.
        Supersedes the parameters in params if provided.
        Default to `False`.

    Attributes
    ----------
    n: float
        The aspect ratio of the domain.
    oceanic_basis: SymbolicBasis
        Object holding the symbolic modes of the ocean.
    atmospheric_basis: None or SymbolicBasis
        Object holding the symbolic modes of the atmosphere (or `None` if there is no atmosphere).
    connected_to_atmosphere: bool
        Indicate if the ocean is connected to an atmosphere.
    stored: bool
        Indicate if the inner product must be stored or computed on the fly.
    ip: InnerProductDefinition
        Object defining the inner product.
    iip: InnerProductDefinition
        Object defining the interaction inner product.
    subs: list(tuple)
        List of 2-tuples containing the substitutions to be made with the functions after the inner products
        symbolic computation.
    """
    def __init__(self, params=None, stored=True, inner_product_definition=None, interaction_inner_product_definition=None,
                 num_threads=None, quadrature=True, timeout=None, dynTinnerproducts=None, T4innerproducts=None, return_symbolic=False, make_substitution=True):

        OceanicInnerProducts.__init__(self)

        self.quadrature = quadrature
        if quadrature:
            timeout = True

        if params is not None:
            if isinstance(params, QgParams):
                self.n = params.scale_params.n
                self.oceanic_basis = params.oceanic_basis
                atm_basis = params.atmospheric_basis
                if T4innerproducts is not None:
                    self._T4 = T4innerproducts
                else:
                    self._T4 = params.T4
                if dynTinnerproducts is not None:
                    self._dynamic_T = dynTinnerproducts
                else:
                    self._dynamic_T = params.dynamic_T
            else:
                self.n = params[0]
                self.oceanic_basis = params[1]
                atm_basis = params[2]
                if T4innerproducts is not None:
                    self._T4 = T4innerproducts
                else:
                    self._T4 = False
                if dynTinnerproducts is not None:
                    self._dynamic_T = dynTinnerproducts
                else:
                    self._dynamic_T = False
        else:
            self.n = None
            self.oceanic_basis = None
            atm_basis = None
            stored = False
            if T4innerproducts is not None:
                self._T4 = T4innerproducts
            else:
                self._T4 = False
            if dynTinnerproducts is not None:
                self._dynamic_T = dynTinnerproducts
            else:
                self._dynamic_T = False

        self.atmospheric_basis = None
        self.connected_to_atmosphere = False

        self.return_symbolic = return_symbolic
        if return_symbolic:
            self.mk_subs = make_substitution
            if self.mk_subs:
                self.subs = [(self.n.symbol, self.n)]
            else:
                self.subs = None
        else:
            self.mk_subs = True
            self.subs = [(self.n.symbol, self.n)]

        if inner_product_definition is None:
            self.ip = StandardSymbolicInnerProductDefinition()
        else:
            self.ip = inner_product_definition

        if interaction_inner_product_definition is None:
            self.iip = self.ip
        else:
            self.iip = interaction_inner_product_definition

        self.stored = stored
        if stored:
            self.compute_inner_products(num_threads, timeout)

        if atm_basis is not None:
            self.connect_to_atmosphere(atm_basis, num_threads, timeout)

    def _F(self, i):
        if self.atmospheric_basis is not None:
            return self.atmospheric_basis[i]

    def _phi(self, i):
        if self.oceanic_basis is not None:
            return self.oceanic_basis[i]

    def connect_to_atmosphere(self, atmosphere_basis, num_threads=None, timeout=None):
        """Connect the ocean to an atmosphere.

        Parameters
        ----------
        atmosphere_basis: SymbolicBasis or AtmosphericSymbolicInnerProducts
            Basis of function of the atmosphere or a symbolic atmospheric inner products object containing the basis.
        num_threads: int or None, optional
            Number of threads to use to compute the symbolic inner products. If `None` use all the cpus available.
            Default to `None`.
        timeout: int or float or bool or None, optional
            The timeout for the computation of each inner product. After the timeout, compute the inner product with a quadrature instead of symbolic integration.
            If `True`, force the timeout and compute directly the inner product with a quadrature instead of trying to do the integration symbolically.
            If `None` or `False`, no timeout occurs.
            Default to `None`.
        """

        if isinstance(atmosphere_basis, AtmosphericSymbolicInnerProducts):
            atmosphere_basis = atmosphere_basis.atmospheric_basis
        self.atmospheric_basis = atmosphere_basis
        self.connected_to_atmosphere = True

        if self.stored:
            if num_threads is None:
                num_threads = cpu_count()

            with Pool(max_workers=num_threads) as pool:
                if self.mk_subs:
                    subs = self.subs + self.oceanic_basis.substitutions + self.atmospheric_basis.substitutions
                else:
                    subs = self.subs

                natm = len(atmosphere_basis)
                if self.return_symbolic:
                    self._K = None
                    self._W = None
                    self._Z = None
                else:
                    self._K = sp.zeros((self.noc, natm), dtype=float, format='dok')
                    self._W = sp.zeros((self.noc, natm), dtype=float, format='dok')
                    if self._T4 or self._dynamic_T:
                        self._Z = sp.zeros((self.noc, natm, natm, natm, natm), dtype=float, format='dok')

                # K inner products
                args_list = [[(i, j), self.iip.ip_lap, (self._phi(i), self._F(j))] for i in range(self.noc)
                             for j in range(natm)]

                output = _parallel_compute(pool, args_list, subs, self._K, timeout, symbolic_int=not self.mk_subs)

                if self.return_symbolic:
                    self._K = ImmutableSparseMatrix(self.noc, natm, output)
                else:
                    self._K = self._K.to_coo()

                # W inner products
                args_list = [[(i, j), self.iip.symbolic_inner_product, (self._phi(i), self._F(j))] for i in range(self.noc)
                             for j in range(natm)]

                output = _parallel_compute(pool, args_list, subs, self._W, timeout, symbolic_int=not self.mk_subs)

                if self.return_symbolic:
                    self._W = ImmutableSparseMatrix(self.noc, natm, output)
                else:
                    self._W = self._W.to_coo()

                if self._T4:
                    # Z inner products
                    args_list = [[(i, j, k, ell, m), self.ip.symbolic_inner_product, (self._phi(i), self._F(j) * self._F(k) * self._F(ell) * self._F(m))] for i in range(self.noc)
                                 for j in range(natm) for k in range(j, natm) for ell in range(k, natm) for m in range(ell, natm)]

                    output = _parallel_compute(pool, args_list, subs, self._Z, timeout, permute=True, symbolic_int=not self.mk_subs)
                elif self._dynamic_T:
                    # Z inner products
                    args_list = [[(i, 0, 0, 0, m), self.ip.symbolic_inner_product, (self._phi(i), self._F(0) * self._F(0) * self._F(0) * self._F(m))]
                                 for i in range(self.noc) for m in range(natm)]

                    output = _parallel_compute(pool, args_list, subs, self._Z, timeout, permute=True, symbolic_int=not self.mk_subs)

            if self._T4 or self._dynamic_T:
                if self.return_symbolic:
                    self._Z = ImmutableSparseNDimArray(output, shape=(self.noc, natm, natm, natm, natm))
                else:
                    self._Z = self._Z.to_coo()

    def compute_inner_products(self, num_threads=None, timeout=None):
        """Function computing and storing all the inner products at once.

        Parameters
        ----------
        num_threads: int or None, optional
            Number of threads to use to compute the symbolic inner products. If `None` use all the cpus available.
            Default to `None`.
        timeout: int or float or bool or None, optional
            The timeout for the computation of each inner product. After the timeout, compute the inner product with a quadrature instead of symbolic integration.
            If `True`, force the timeout and compute directly the inner product with a quadrature instead of trying to do the integration symbolically.
            If `None` or `False`, no timeout occurs.
            Default to `None`.
        """
        if self.return_symbolic:
            self._M = None
            self._U = None
            self._N = None
            self._O = None
            self._C = None
            self._V = None
        else:
            self._M = sp.zeros((self.noc, self.noc), dtype=float, format='dok')
            self._U = sp.zeros((self.noc, self.noc), dtype=float, format='dok')
            self._N = sp.zeros((self.noc, self.noc), dtype=float, format='dok')
            self._O = sp.zeros((self.noc, self.noc, self.noc), dtype=float, format='dok')
            self._C = sp.zeros((self.noc, self.noc, self.noc), dtype=float, format='dok')
            if self._T4 or self._dynamic_T:
                self._V = sp.zeros((self.noc, self.noc, self.noc, self.noc, self.noc), dtype=float, format='dok')

        if self.stored:
            if num_threads is None:
                num_threads = cpu_count()

            with Pool(max_workers=num_threads) as pool:
                if self.mk_subs:
                    subs = self.subs + self.oceanic_basis.substitutions
                else:
                    subs = self.subs

                # N inner products
                args_list = [[(i, j), self.ip.ip_diff_x, (self._phi(i), self._phi(j))] for i in range(self.noc) for j in range(self.noc)]

                output = _parallel_compute(pool, args_list, subs, self._N, timeout, symbolic_int=not self.mk_subs)
                if self.return_symbolic:
                    self._N = ImmutableSparseMatrix(self.noc, self.noc, output)
                else:
                    self._N = self._N.to_coo()

                # M inner products
                args_list = [[(i, j), self.ip.ip_lap, (self._phi(i), self._phi(j))] for i in range(self.noc) for j in range(self.noc)]

                output = _parallel_compute(pool, args_list, subs, self._M, timeout, symbolic_int=not self.mk_subs)
                if self.return_symbolic:
                    self._M = ImmutableSparseMatrix(self.noc, self.noc, output)
                else:
                    self._M = self._M.to_coo()

                # U inner products
                args_list = [[(i, j), self.ip.symbolic_inner_product, (self._phi(i), self._phi(j))] for i in range(self.noc) for j in range(self.noc)]

                output = _parallel_compute(pool, args_list, subs, self._U, timeout, symbolic_int=not self.mk_subs)
                if self.return_symbolic:
                    self._U = ImmutableSparseMatrix(self.noc, self.noc, output)
                else:
                    self._U = self._U.to_coo()

                # O inner products
                args_list = [[(i, j, k), self.ip.ip_jac, (self._phi(i), self._phi(j), self._phi(k))] for i in range(self.noc)
                             for j in range(self.noc) for k in range(self.noc)]

                output = _parallel_compute(pool, args_list, subs, self._O, timeout, symbolic_int=not self.mk_subs)
                if self.return_symbolic:
                    self._O = ImmutableSparseNDimArray(output, shape=(self.noc, self.noc, self.noc))
                else:
                    self._O = self._O.to_coo()

                # C inner products
                args_list = [[(i, j, k), self.ip.ip_jac_lap, (self._phi(i), self._phi(j), self._phi(k))] for i in range(self.noc)
                             for j in range(self.noc) for k in range(self.noc)]

                output = _parallel_compute(pool, args_list, subs, self._C, timeout, symbolic_int=not self.mk_subs)
                if self.return_symbolic:
                    self._C = ImmutableSparseNDimArray(output, shape=(self.noc, self.noc, self.noc))
                else:
                    self._C = self._C.to_coo()

                if self._T4:
                    # V inner products
                    args_list = [[(i, j, k, ell, m), self.ip.symbolic_inner_product, (self._phi(i), self._phi(j) * self._phi(k) * self._phi(ell) * self._phi(m))] for i in range(self.noc)
                                 for j in range(self.noc) for k in range(j, self.noc) for ell in range(k, self.noc) for m in range(ell, self.noc)]

                    output = _parallel_compute(pool, args_list, subs, self._V, timeout, permute=True, symbolic_int=not self.mk_subs)
                elif self._dynamic_T:
                    # V inner products
                    args_list = [[(i, 0, 0, 0, m), self.ip.symbolic_inner_product, (self._phi(i), self._phi(0) * self._phi(0) * self._phi(0) * self._phi(m))]
                                 for i in range(self.noc) for m in range(self.noc)]

                    output = _parallel_compute(pool, args_list, subs, self._V, timeout, permute=True, symbolic_int=not self.mk_subs)

                if self._T4 or self._dynamic_T:
                    if self.return_symbolic:
                        self._V = ImmutableSparseNDimArray(output, shape=(self.noc, self.noc, self.noc, self.noc, self.noc))
                    else:
                        self._V = self._V.to_coo()
            
    @property
    def noc(self):
        """Number of oceanic modes."""
        return len(self.oceanic_basis)

    # !-----------------------------------------------------!
    # ! Inner products in the equations for the ocean       !
    # !-----------------------------------------------------!

    def _integrate(self, subs, args):
        if self.return_symbolic:
            res = _apply(args)
            return res[1]
        
        if self.quadrature:
            res = _num_apply(args)
            return res[1]
        else:
            res = _apply(args)[1]
            return float(res.subs(subs))

    def M(self, i, j):
        """Function to compute the forcing of the ocean fields on the ocean: :math:`M_{i,j} = (\\phi_i, \\nabla^2 \\phi_j)`."""
        if not self.stored:
            if self.mk_subs:
                subs = self.subs + self.oceanic_basis.substitutions
            else:
                subs = self.subs

            args = ((i, j), self.ip.ip_lap, (self._phi(i), self._phi(j)), subs)
            return self._integrate(subs, args)
        else:
            return self._M[i, j]

    def U(self, i, j):
        """Function to compute the inner products: :math:`U_{i,j} = (\\phi_i, \\phi_j)`."""
        if not self.stored:
            if self.mk_subs:
                subs = self.subs + self.oceanic_basis.substitutions
            else:
                subs = self.subs

            args = ((i, j), self.ip.symbolic_inner_product, (self._phi(i), self._phi(j)), subs)
            return self._integrate(subs, args)
        else:
            return self._U[i, j]

    def N(self, i, j):
        """Function computing the beta term for the ocean: :math:`N_{i,j} = (\\phi_i, \\partial_x \\phi_j)`."""
        if not self.stored:
            if self.mk_subs:
                subs = self.subs + self.oceanic_basis.substitutions
            else:
                subs = self.subs

            args = ((i, j), self.ip.ip_diff_x, (self._phi(i), self._phi(j)), subs)
            return self._integrate(subs, args)
        else:
            return self._N[i, j]

    def O(self, i, j, k):
        """Function to compute the temperature advection term (passive scalar): :math:`O_{i,j,k} = (\\phi_i, J(\\phi_j, \\phi_k))`"""
        if not self.stored:
            if self.mk_subs:
                subs = self.subs + self.oceanic_basis.substitutions
            else:
                subs = self.subs
            
            args = ((i, j, k), self.ip.ip_jac, (self._phi(i), self._phi(j), self._phi(k)), subs)
            return self._integrate(subs, args)
        else:
            return self._O[i, j, k]

    def C(self, i, j, k):
        """Function to compute the tensors holding the Jacobian inner products: :math:`C_{i,j,k} = (\\phi_i, J(\\phi_j,\\nabla^2 \\phi_k))`."""
        if not self.stored:
            if self.mk_subs:
                subs = self.subs + self.oceanic_basis.substitutions
            else:
                subs = self.subs

            args = ((i, j, k), self.ip.ip_jac_lap, (self._phi(i), self._phi(j), self._phi(k)), subs)
            return self._integrate(subs, args)
        else:
            return self._C[i, j, k]

    def K(self, i, j):
        """Function to commpute the forcing of the ocean by the atmosphere: :math:`K_{i,j} = (\\phi_i, \\nabla^2 F_j)`."""
        if self.connected_to_atmosphere:
            if not self.stored:
                if self.mk_subs:
                    subs = self.subs + self.oceanic_basis.substitutions + self.atmospheric_basis.substitutions
                else:
                    subs = self.subs

                args = ((i, j), self.iip.ip_lap, (self._phi(i), self._F(j)), subs)
                return self._integrate(subs, args)
            else:
                return self._K[i, j]
        else:
            return 0

    def W(self, i, j):
        """Function to compute the short-wave radiative forcing of the ocean: :math:`W_{i,j} = (\\phi_i, F_j)`."""
        if self.connected_to_atmosphere:
            if not self.stored:
                if self.mk_subs:
                    subs = self.subs + self.oceanic_basis.substitutions + self.atmospheric_basis.substitutions
                else:
                    subs = self.subs

                args = ((i, j), self.iip.symbolic_inner_product, (self._phi(i), self._F(j)), subs)
                return self._integrate(subs, args)
            else:
                return self._W[i, j]
        else:
            return 0

    def V(self, i, j, k, l, m):
        """Function to compute the :math:`T^4` temperature forcing from the ocean to the atmosphere: :math:`V_{i,j,k,l,m} = (\\phi_i, \\phi_j, \\phi_k, \\phi_l, \\phi_m)`."""
        if not self.stored:
            if self.mk_subs:
                subs = self.subs + self.oceanic_basis.substitutions
            else:
                subs = self.subs

            args = ((i, j, k, l, m), self.ip.symbolic_inner_product, (self._phi(i), self._phi(j) * self._phi(k) * self._phi(l) * self._phi(m)), subs)
            return self._integrate(subs, args)
        else:
            return self._V[i, j, k, l, m]

    def Z(self, i, j, k, l, m):
        """Function to compute the :math:`T^4` temperature forcing from the atmosphere to the ocean: :math:`Z_{i,j,k,l,m} = (\\phi_i, F_j, F_k, F_l, F_m)`."""
        if not self.stored:
            if self.mk_subs:
                subs = self.subs + self.oceanic_basis.substitutions + self.atmospheric_basis.substitutions
            else:
                subs = self.subs

            args = ((i, j, k, l, m), self.ip.symbolic_inner_product, (self._phi(i), self._F(j) * self._F(k) * self._F(l) * self._F(m)), subs)
            return self._integrate(subs, args)
        else:
            return self._Z[i, j, k, l, m]


class GroundSymbolicInnerProducts(GroundInnerProducts):
    """Class which contains all the ground inner products coefficients needed for the tendencies
    tensor :class:`~.tensors.qgtensor.QgsTensor` computation.

    Parameters
    ----------
    params: QgParams or list
        An instance of model's parameters object or a list in the form [aspect_ratio, ground_basis, atmospheric_basis].
        If a list is provided, `aspect_ratio` is the aspect ratio of the domain, `ground_basis` is a SymbolicBasis object
        with the modes of the ground, and `atmospheric_basis` is either a SymbolicBasis with the modes of the atmosphere
        or `None` if there is no atmosphere.
    stored: bool, optional
        Indicate if the inner product must be stored or computed on the fly. Default to `True`
    inner_product_definition: None or InnerProductDefinition, optional
        The definition of the inner product being used. If `None`, use the canonical StandardInnerProductDefinition object.
        Default to `None`.
    interaction_inner_product_definition: None or InnerProductDefinition, optional
        The definition of the inner product being used for the interaction with the other components, i.e. to compute the inner products with the other component base of funcitons.
        If `None`, use the `inner_product_definition` provided.
        Default to `None`.
    num_threads: int or None, optional
        Number of threads to use to compute the symbolic inner products. If `None` use all the cpus available.
        Default to `None`.
    quadrature: bool, optional
        If `True', compute the inner products with a quadrature instead of the symbolic integration.
        If `True` Disable the `timeout` parameter.
        Default to `True`.
    timeout: int or float or bool or None, optional
        The timeout for the computation of each inner product. After the timeout, compute the inner product with a quadrature instead of symbolic integration.
        If `None` or `False`, no timeout occurs.
        Default to `None`.
    dynTinnerproducts: bool, optional
        If the inner products are stored, allow to compute or not the inner products corresponding to the dynamic temperature tendencies.
        Supersedes the parameters in params if provided.
        Default to `False`.
    T4innerproducts: bool, optional
        If the inner products are stored, allow to compute or not the inner products corresponding to the :math:`T^4` tendencies.
        Compute the inner products corresponding to the dynamic temperature tendencies as well.
        Default to `False`.

    Attributes
    ----------
    n: float
        The aspect ratio of the domain.
    ground_basis: SymbolicBasis
        Object holding the symbolic modes of the ground.
    atmospheric_basis: None or SymbolicBasis
        Object holding the symbolic modes of the atmosphere (or `None` if there is no atmosphere).
    connected_to_atmosphere: bool
        Indicate if the ground is connected to an atmosphere.
    stored: bool
        Indicate if the inner product must be stored or computed on the fly.
    ip: InnerProductDefinition
        Object defining the inner product.
    iip: InnerProductDefinition
        Object defining the interaction inner product.
    subs: list(tuple)
        List of 2-tuples containing the substitutions to be made with the functions after the inner products
        symbolic computation.
    """

    def __init__(self, params=None, stored=True, inner_product_definition=None, interaction_inner_product_definition=None,
                 num_threads=None, quadrature=True, timeout=None, dynTinnerproducts=None, T4innerproducts=None, return_symbolic=False, make_substitution=True):

        GroundInnerProducts.__init__(self)

        self.quadrature = quadrature
        if quadrature:
            timeout = True

        if params is not None:
            if isinstance(params, QgParams):
                self.n = params.scale_params.n
                self.ground_basis = params.ground_basis
                atm_basis = params.atmospheric_basis
                if T4innerproducts is not None:
                    self._T4 = T4innerproducts
                else:
                    self._T4 = params.T4
                if dynTinnerproducts is not None:
                    self._dynamic_T = dynTinnerproducts
                else:
                    self._dynamic_T = params.dynamic_T
            else:
                self.n = params[0]
                self.ground_basis = params[1]
                atm_basis = params[2]
                if T4innerproducts is not None:
                    self._T4 = T4innerproducts
                else:
                    self._T4 = False
                if dynTinnerproducts is not None:
                    self._dynamic_T = dynTinnerproducts
                else:
                    self._dynamic_T = False
        else:
            self.n = None
            self.ground_basis = None
            atm_basis = None
            stored = False
            if T4innerproducts is not None:
                self._T4 = T4innerproducts
            else:
                self._T4 = False
            if dynTinnerproducts is not None:
                self._dynamic_T = dynTinnerproducts
            else:
                self._dynamic_T = False

        self.atmospheric_basis = None
        self.connected_to_atmosphere = False

        self.return_symbolic = return_symbolic
        if return_symbolic:
            self.mk_subs = make_substitution
            if self.mk_subs:
                self.subs = [(self.n.symbol, self.n)]
            else:
                self.subs = None
        else:
            self.mk_subs = True
            self.subs = [(self.n.symbol, self.n)]

        if inner_product_definition is None:
            self.ip = StandardSymbolicInnerProductDefinition()
        else:
            self.ip = inner_product_definition

        if interaction_inner_product_definition is None:
            self.iip = self.ip
        else:
            self.iip = interaction_inner_product_definition

        self.stored = stored
        if stored:
            self.compute_inner_products(num_threads, timeout)

        if atm_basis is not None:
            self.connect_to_atmosphere(atm_basis, num_threads, timeout)

    def _F(self, i):
        if self.atmospheric_basis is not None:
            return self.atmospheric_basis[i]

    def _phi(self, i):
        if self.ground_basis is not None:
            return self.ground_basis[i]

    def connect_to_atmosphere(self, atmosphere_basis, num_threads=None, timeout=None):
        """Connect the ocean to an atmosphere.

        Parameters
        ----------
        atmosphere_basis: SymbolicBasis or AtmosphericSymbolicInnerProducts
            Basis of function of the atmosphere or a symbolic atmospheric inner products object containing the basis.
        num_threads: int or None, optional
            Number of threads to use to compute the symbolic inner products. If `None` use all the cpus available.
            Default to `None`.
        timeout: int or float or bool or None, optional
            The timeout for the computation of each inner product. After the timeout, compute the inner product with a quadrature instead of symbolic integration.
            If `True`, force the timeout and compute directly the inner product with a quadrature instead of trying to do the integration symbolically.
            If `None` or `False`, no timeout occurs.
            Default to `None`.
        """
        if isinstance(atmosphere_basis, AtmosphericSymbolicInnerProducts):
            atmosphere_basis = atmosphere_basis.atmospheric_basis
        self.atmospheric_basis = atmosphere_basis
        self.connected_to_atmosphere = True

        if self.stored:
            if num_threads is None:
                num_threads = cpu_count()

            with Pool(max_workers=num_threads) as pool:
                if self.mk_subs:
                    subs = self.subs + self.ground_basis.substitutions + self.atmospheric_basis.substitutions
                else:
                    subs = self.subs

                natm = len(atmosphere_basis)
                if self.return_symbolic:
                    self._W = None
                    self._Z = None
                else:
                    self._W = sp.zeros((self.ngr, natm), dtype=float, format='dok')
                    if self._T4 or self._dynamic_T:
                        self._Z = sp.zeros((self.ngr, natm, natm, natm, natm), dtype=float, format='dok')

                # W inner products
                args_list = [[(i, j), self.iip.symbolic_inner_product, (self._phi(i), self._F(j))] for i in range(self.ngr)
                             for j in range(natm)]
                output = _parallel_compute(pool, args_list, subs, self._W, timeout, symbolic_int=not self.mk_subs)
                if self.return_symbolic:
                    self._W = ImmutableSparseMatrix(self.ngr, natm, output)
                else:
                    self._W = self._W.to_coo()

                if self._T4:
                    # Z inner products
                    args_list = [[(i, j, k, ell, m), self.ip.symbolic_inner_product, (self._phi(i), self._F(j) * self._F(k) * self._F(ell) * self._F(m))] for i in range(self.ngr)
                                 for j in range(natm) for k in range(j, natm) for ell in range(k, natm) for m in range(ell, natm)]

                    output = _parallel_compute(pool, args_list, subs, self._Z, timeout, permute=True, symbolic_int=not self.mk_subs)
                elif self._dynamic_T:
                    # Z inner products
                    args_list = [[(i, 0, 0, 0, m), self.ip.symbolic_inner_product, (self._phi(i), self._F(0) * self._F(0) * self._F(0) * self._F(m))]
                                 for i in range(self.ngr) for m in range(natm)]

                    output = _parallel_compute(pool, args_list, subs, self._Z, timeout, permute=True, symbolic_int=not self.mk_subs)

            if self._T4 or self._dynamic_T:
                if self.return_symbolic:
                    self._Z = ImmutableSparseNDimArray(output, shape=(self.ngr, natm, natm, natm, natm))
                else:
                    self._Z = self._Z.to_coo()

    def compute_inner_products(self, num_threads=None, timeout=None):
        """Function computing and storing all the inner products at once.

        Parameters
        ----------
        num_threads: int or None, optional
            Number of threads to use to compute the symbolic inner products. If `None` use all the cpus available.
            Default to `None`.
        timeout: int or float or bool or None, optional
            The timeout for the computation of each inner product. After the timeout, compute the inner product with a quadrature instead of symbolic integration.
            If `True`, force the timeout and compute directly the inner product with a quadrature instead of trying to do the integration symbolically.
            If `None` or `False`, no timeout occurs.
            Default to `None`.
        """
        if self.return_symbolic:
            self._U = None
            self._V = None
        else:
            self._U = sp.zeros((self.ngr, self.ngr), dtype=float, format='dok')
            if self._T4 or self._dynamic_T:
                self._V = sp.zeros((self.ngr, self.ngr, self.ngr, self.ngr, self.ngr), dtype=float, format='dok')

        if self.stored:
            if num_threads is None:
                num_threads = cpu_count()

            with Pool(max_workers=num_threads) as pool:
                if self.mk_subs:
                    subs = self.subs + self.ground_basis.substitutions
                else:
                    subs = self.subs

                # U inner products
                args_list = [[(i, j), self.ip.symbolic_inner_product, (self._phi(i), self._phi(j))] for i in range(self.ngr) for j in range(self.ngr)]
                output = _parallel_compute(pool, args_list, subs, self._U, timeout, symbolic_int=not self.mk_subs)
                if self.return_symbolic:
                    self._U = ImmutableSparseMatrix(self.ngr, self.ngr, output)
                else:
                    self._U = self._U.to_coo()

                if self._T4:
                    # V inner products
                    args_list = [[(i, j, k, ell, m), self.ip.symbolic_inner_product, (self._phi(i), self._phi(j) * self._phi(k) * self._phi(ell) * self._phi(m))] for i in range(self.ngr)
                                 for j in range(self.ngr) for k in range(j, self.ngr) for ell in range(k, self.ngr) for m in range(ell, self.ngr)]

                    output = _parallel_compute(pool, args_list, subs, self._V, timeout, permute=True, symbolic_int=not self.mk_subs)

                elif self._dynamic_T:
                    # V inner products
                    args_list = [[(i, 0, 0, 0, m), self.ip.symbolic_inner_product, (self._phi(i), self._phi(0) * self._phi(0) * self._phi(0) * self._phi(m))]
                                 for i in range(self.ngr) for m in range(self.ngr)]

                    output = _parallel_compute(pool, args_list, subs, self._V, timeout, permute=True, symbolic_int=not self.mk_subs)

            if self._T4 or self._dynamic_T:
                if self.return_symbolic:
                    self._V = ImmutableSparseNDimArray(output, shape=(self.ngr, self.ngr, self.ngr, self.ngr, self.ngr))
                else:
                    self._V = self._V.to_coo()

    @property
    def ngr(self):
        """Number of ground modes."""
        return len(self.ground_basis)

    # !-----------------------------------------------------!
    # ! Inner products in the equations for the ocean       !
    # !-----------------------------------------------------!

    def _integrate(self, subs, args):
        if self.return_symbolic:
            res = _apply(args)
            return res[1]
        
        if self.quadrature:
            res = _num_apply(args)
            return res[1]
        else:
            res = _apply(args)[1]
            return float(res.subs(subs))

    def K(self, i, j):
        """Function to commpute the forcing: :math:`K_{i,j} = (\\phi_i, \\nabla^2 F_j)`.

        Warnings
        --------
        Not defined and not used.
        """
        return 0

    def M(self, i, j):
        """Function to compute the forcing: :math:`M_{i,j} = (\\phi_i, \\nabla^2 \\phi_j)`.

        Warnings
        --------
        Not defined and not used.
        """
        return 0

    def U(self, i, j):
        """Function to compute the inner products: :math:`U_{i,j} = (\\phi_i, \\phi_j)`."""
        if not self.stored:
            if self.mk_subs:
                subs = self.subs + self.ground_basis.substitutions
            else:
                subs = self.subs

            args = ((i, j), self.ip.symbolic_inner_product, (self._phi(i), self._phi(j)), subs)
            return self._integrate(subs, args)
        else:
            return self._U[i, j]

    def N(self, i, j):
        """Function computing the beta term: :math:`N_{i,j} = (\\phi_i, \\partial_x \\phi_j)`.

        Warnings
        --------
        Not defined and not used.
        """
        return 0

    def O(self, i, j, k):
        """Function to compute the temperature advection term (passive scalar): :math:`O_{i,j,k} = (\\phi_i, J(\\phi_j, \\phi_k))`

        Warnings
        --------
        Not defined and not used.
        """
        return 0

    def C(self, i, j, k):
        """Function to compute the tensors holding the Jacobian inner products: :math:`C_{i,j,k} = (\\phi_i, J(\\phi_j,\\nabla^2 \\phi_k))`.

        Warnings
        --------
        Not defined and not used.
        """
        return 0

    def W(self, i, j):
        """Function to compute the short-wave radiative forcing of the ground: :math:`W_{i,j} = (\\phi_i, F_j)`."""
        if self.connected_to_atmosphere:
            if not self.stored:
                if self.mk_subs:
                    subs = self.subs + self.ground_basis.substitutions + self.atmospheric_basis.substitutions
                else:
                    subs = self.subs

                args = ((i, j), self.iip.symbolic_inner_product, (self._phi(i), self._F(j)), subs)
                return self._integrate(subs, args)
            else:
                return self._W[i, j]
        else:
            return 0

    def V(self, i, j, k, l, m):
        """Function to compute the :math:`T^4` temperature forcing from the ground to the atmosphere: :math:`V_{i,j,k,l,m} = (\\phi_i, \\phi_j, \\phi_k, \\phi_l, \\phi_m)`."""
        if not self.stored:
            if self.mk_subs:
                subs = self.subs + self.ground_basis.substitutions
            else:
                subs = self.subs

            args = ((i, j, k, l, m), self.ip.symbolic_inner_product, (self._phi(i), self._phi(j) * self._phi(k) * self._phi(l) * self._phi(m)), subs)
            return self._integrate(subs, args)
        else:
            return self._V[i, j, k, l, m]

    def Z(self, i, j, k, l, m):
        """Function to compute the :math:`T^4` temperature forcing from the atmosphere to the ground: :math:`Z_{i,j,k,l,m} = (\\phi_i, F_j, F_k, F_l, F_m)`."""
        if not self.stored:
            if self.mk_subs:
                subs = self.subs + self.ground_basis.substitutions + self.atmospheric_basis.substitutions
            else:
                subs = self.subs
                
            args = ((i, j, k, l, m), self.ip.symbolic_inner_product, (self._phi(i), self._F(j) * self._F(k) * self._F(l) * self._F(m)), subs)
            return self._integrate(subs, args)
        else:
            return self._Z[i, j, k, l, m]


def _apply(ls):
    return ls[0], ls[1](*ls[2])


def _num_apply(ls):
    integrand = ls[1](*ls[2], integrand=True)

    if ls[3] is not None:
        num_integrand = integrand[0].subs(ls[3])
    else:
        num_integrand = integrand[0]

    func = lambdify((integrand[1][0], integrand[2][0]), num_integrand, 'numpy')

    try:
        a = integrand[2][1].subs(ls[3])
    except:
        a = integrand[2][1]
    try:
        a = a.evalf()
    except:
        pass
    try:
        b = integrand[2][2].subs(ls[3])
    except:
        b = integrand[2][2]
    try:
        b = b.evalf()
    except:
        pass
    try:
        gfun = integrand[1][1].subs(ls[3])
    except:
        gfun = integrand[1][1]
    try:
        gfun = gfun.evalf()
    except:
        pass
    try:
        hfun = integrand[1][2].subs(ls[3])
    except:
        hfun = integrand[1][2]
    try:
        hfun = hfun.evalf()
    except:
        pass

    res = dblquad(func, a, b, gfun, hfun)
  
    if abs(res[0]) <= res[1]:
        return ls[0], 0
    else:
        return ls[0], res[0]


def _parallel_compute(pool, args_list, subs, destination, timeout, permute=False, symbolic_int=False):
    if destination is None:
        return_dict = True
        destination = dict()
    else:
        return_dict = False

    if timeout is False or symbolic_int:
        timeout = None

    if timeout is not True:
        future = pool.map(_apply, args_list, timeout=timeout)
        results = future.result()
        num_args_list = list()
        i = 0
        while True:
            try:
                res = next(results)
                if symbolic_int:
                    expr = res[1].simplify()
                    destination[res[0]] = expr
                    if permute:
                        i = res[0][0]
                        idx = res[0][1:]
                        perm_idx = multiset_permutations(idx)
                        for perm in perm_idx:
                            idx = [i] + perm
                            destination[tuple(idx)] = expr
                else:
                    destination[res[0]] = float(res[1].subs(subs))
            except StopIteration:
                break
            except TimeoutError:
                num_args_list.append(args_list[i] + [subs])
            i += 1
    else:
        num_args_list = [args + [subs] for args in args_list]

    future = pool.map(_num_apply, num_args_list)
    results = future.result()
    if permute:
        while True:
            try:
                res = next(results)
                i = res[0][0]
                idx = res[0][1:]
                perm_idx = multiset_permutations(idx)
                for perm in perm_idx:
                    idx = [i] + perm
                    destination[tuple(idx)] = res[1]
            except StopIteration:
                break
    else:
        while True:
            try:
                res = next(results)
                destination[res[0]] = res[1]
            except StopIteration:
                break

    if return_dict:
        return destination


if __name__ == '__main__':
    from qgs.params.params import QgParams
    pars = QgParams(dynamic_T=True)  # , T4=True)
    pars.set_atmospheric_channel_fourier_modes(2, 2, mode='symbolic')
    pars.set_oceanic_basin_fourier_modes(2, 4, mode='symbolic')
    # aip = AtmosphericSymbolicInnerProducts(pars, quadrature=True)
    # oip = OceanicSymbolicInnerProducts(pars, quadrature=True)
    aip = AtmosphericSymbolicInnerProducts(pars, quadrature=True, T4innerproducts=True)
    oip = OceanicSymbolicInnerProducts(pars, quadrature=True, T4innerproducts=True)
